/***************************************************************************
****************************************************************************
****************************************************************************
*
* Clobberd - By Jason Nunn
* FREEWARE. Authorship Reserved 1996
*
* "A user regulator with attitude.." (ANSI C)
*
* Snail: 32 Rothdale Road, Moil, Darwin, NT, 0810, Australia
*
* ==================================================================
*
* Developed on a Linux 1.2.13 box ...then later 2.0.0
*
****************************************************************************
****************************************************************************
***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <time.h>
#include <utmp.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <pwd.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <dirent.h>
#include <syslog.h>
#include "clobberd.h"
#include "misc.h"

int           daemon_mode = 1;
int           running = 1;
int           listen_loop;
enum
{
  MONITOR,
  EDIT,
  RELOAD
} edit_mode = MONITOR;
int           midnight_reset_f = 0;
int           total_purge_f = 0;

tusettings_hr usettings_hr = DEFAULT_HR;
int           usettings_list_size = 0;
tusettings    **usettings_list = NULL;
tutmp_hr      utmp_hr = {0};
tutmp_ll      utmp_ll_root = {NULL,NULL};

time_t        timer,old_timer;

int           old_total_jiffies;
int           total_jiffies;
unsigned int  total_rss;
int           no_dialins_free;

int            newaddrsize;
int            main_socket;
char           our_hostname[STR_LEN];

/***************************************************************************
*
***************************************************************************/
void log_msg(char *msg,int rl)
{
  if(usettings_hr.run_level <= rl)
    if(daemon_mode)
    {
      openlog("clobberd",LOG_CONS,LOG_DAEMON);
      syslog(LOG_NOTICE,msg);
      closelog();
    }
    else
      printf("%s\n",msg);
}

/***************************************************************************
*
***************************************************************************/
void kill_daemon(int sig)
{
  running = 0;
  listen_loop = 0;
}

void set_editmode(int sig)
{
  signal(sig,set_editmode);
  edit_mode = EDIT;
  log_msg("S: Entering Edit mode. Safe to edit files",1);
}

void set_monitormode(int sig)
{
  signal(sig,set_monitormode);
  edit_mode = RELOAD;
  log_msg("S: Exiting Edit mode. Clobberd will reload files next wakeup",1);
}

void set_midnightreset(int sig)
{
  signal(sig,set_midnightreset);
  midnight_reset_f = 1;
  log_msg("S: Force Midnight reset.",1);
}

void set_us_purge(int sig)
{
  signal(sig,set_us_purge);
  total_purge_f = 1;
  log_msg("S: Force Total Purge.",1);
}

/***************************************************************************
*
***************************************************************************/
void init(void)
{
  register int x;

  newaddrsize = sizeof(struct sockaddr_in);
  for(x = 0;x < NSIG;x++) signal(x,SIG_IGN);
  signal(SIGINT,kill_daemon);
  signal(SIGKILL,kill_daemon);
  signal(SIGTERM,kill_daemon);
  signal(SIGUSR1,set_editmode);
  signal(SIGUSR2,set_monitormode);
  signal(SIGQUIT,set_midnightreset);
  signal(SIGWINCH,set_us_purge);
}

/***************************************************************************
*
***************************************************************************/
void daemonise(void)
{
  if(daemon_mode)
    if(fork() == 0)
    {
      register int x;

      for(x = 0;x <= FD_SETSIZE;++x) close(x);
      setsid();
    }
    else
      exit(0);
}

/***************************************************************************
* Network Shit..
***************************************************************************/
/*routine from "Mail Flash" 1994 CHA0S All Rights Reserved*/
int tcpip_client_creat(char *server,int port)
{
  struct sockaddr_in sin;
  struct hostent *hp;
  register int thesock;
 
  hp = gethostbyname(server);
  if(hp == NULL) return -1;
  memset((char *)&sin,0,sizeof(struct sockaddr_in));
  memcpy((char *)&sin.sin_addr,hp->h_addr,hp->h_length);
  sin.sin_family = hp->h_addrtype;
  sin.sin_port = htons(port);
  thesock = socket(AF_INET,SOCK_STREAM,0);
  if(thesock != -1)
    if(connect(thesock,(struct sockaddr *)&sin,sizeof(sin)) == -1)
      return -1;
  return thesock;
}

int tcpip_serv_creat(void)
{
  int newsock,reuse_opt = 1;
  struct sockaddr_in addr = {0};

  newsock = socket(AF_INET,SOCK_STREAM,0);
  setsockopt(newsock,SOL_SOCKET,SO_REUSEADDR,&reuse_opt,sizeof(reuse_opt));
  addr.sin_family = AF_INET;
  addr.sin_port = htons(NET_PORT);
  addr.sin_addr.s_addr = htonl(INADDR_ANY);
  if(bind(newsock,(struct sockaddr *)&addr,sizeof(addr)) != 0)
  {
    log_msg("E: Bind failed. Server is probably already running on"
           " this machine (tcpip_serv_creat).",1);
    return -1;
  }
  fcntl(newsock,F_SETFL,O_NONBLOCK);
  listen(newsock,MAX_CONNECTIONS);
  return newsock;
}

void net_write_pkt(int socket,void *data,int data_size)
{
  fd_set fds;
  struct timeval tv = {60,0};

  FD_ZERO(&fds);
  FD_SET(socket,&fds);
  if(select(socket + 1,NULL,&fds,NULL,&tv))
    send(socket,data,data_size,0);
}

#define net_printf(s,d) net_write_pkt(s,d,strlen(d))

void tcpip_serv_reaper(void)
{
  shutdown(main_socket,2);
  close(main_socket);
}

/***************************************************************************
* READ clobberd.cf
***************************************************************************/
int compar_tusettings(const void *entry1,const void *entry2)
{
  return (*(tusettings **)entry1)->uid - (*(tusettings **)entry2)->uid;
}

tusettings *find_user_settings(uid_t uid)
{
  tusettings *key2 = NULL;

  if(usettings_list_size)
  {
    tusettings key;

    key2 = &key;
    key.uid = uid;
    key2 = (tusettings *)bsearch(&key2,usettings_list,
                         usettings_list_size,
                         sizeof(tusettings *),compar_tusettings);
  }
  if(key2 == NULL)
  {
    if(usettings_hr.def_usettings_enab)
      return &(usettings_hr.def_usettings);
    else
      return NULL;
  }
  return *(tusettings **)key2;
}

int get_user_settings(void)
{
  FILE *fp;
  register int no_entries;

  if((fp = fopen(CONFIG_FILE,"rb")) == NULL)
  {
    log_msg("E: Couldn't open " CONFIG_FILE " (get_user_settings)",1);
    return 0;
  }
  fseek(fp,0,SEEK_END);
  no_entries = ftell(fp);
  fseek(fp,0,SEEK_SET);
  if(no_entries > 0)
  {
    if(fread(&usettings_hr,1,sizeof(tusettings_hr),fp) == sizeof(usettings_hr))
    {
      no_entries -= sizeof(tusettings_hr);
      no_entries /= sizeof(tusettings);
      if(no_entries > 0)
      {
        register int c;

        if((usettings_list =
          malloc(no_entries * sizeof(tusettings *))) == NULL)
        {
          log_msg("E: Couldn't malloc 'usettings_list' (load_user_table)",1);
          return 0;
        }
        for(c = 0;c < no_entries;c++)
        {
          tusettings *curr;

          curr = *(usettings_list + usettings_list_size) = malloc(sizeof(tusettings));
          if(curr == NULL)
          {
            log_msg("E: Couldn't malloc 'usettings' entry (load_user_table)",1);
            return 0;
          }
          if(fread(curr,1,sizeof(tusettings),fp) == sizeof(tusettings))
          {
            if(getpwuid(curr->uid) != NULL)
              usettings_list_size++;
            else
            {
              char str[STR_LEN];

              sprintf(str,"I: Invalid uid entry %d in %s file",curr->uid,CONFIG_FILE);
              log_msg(str,1);
              free(curr);
            }
          }
          else
            free(curr);
        }
        qsort(usettings_list,usettings_list_size,sizeof(tusettings *),compar_tusettings);
      }
      else
        log_msg("W: User list empty. Nothing to monitor...",0);
    }
    else
      log_msg("E: " CONFIG_FILE " empty????. Wierd. (get_user_settings)",1);
  }
  fclose(fp);
  return 1;
}

void free_user_settings(void)
{
  register int c;
  if(usettings_list == NULL) return;
  for(c = 0;c < usettings_list_size;c++) free(*(usettings_list + c));
  free(usettings_list);
  usettings_list = NULL;
  usettings_list_size = 0;
}

/***************************************************************************
* Reads /etc/utmp file, and loads new news into tutmp struct..
***************************************************************************/
tutmp_ll *find_user(tutmp_ll *entry,char *user)
{
  tutmp_ll *fentry(tutmp_ll *entry)
  {
    tutmp_ll *x = NULL;

    if(entry->curr != NULL)
      if(strncmp(entry->curr->user,user,UT_NAMESIZE) != 0)
      {
        if(entry->next != NULL)
          x = fentry(entry->next);
      }
      else
        x = entry;
    return x;
  }

  return fentry(entry);
}

int get_active_users(void)
{
  FILE *fp;
  char str[STR_LEN],substr[STR_LEN];
  struct utmp tmp_buf;

  tutmp_ll *set_actives_gotoend(tutmp_ll *entry)
  {
    tutmp_ll *x = NULL;

    if(entry->curr != NULL)
    {
      register int a = entry->curr->active;
      entry->curr->active = 0;
      rewind(fp);
      while(fread(&tmp_buf,sizeof(struct utmp),1,fp) == 1)
        if(!strncmp(entry->curr->user,tmp_buf.ut_user,UT_NAMESIZE) &&
           (tmp_buf.ut_type == USER_PROCESS))
        {
          entry->curr->pid = tmp_buf.ut_pid;
          entry->curr->active = 1;
          if(!a)
          {
            tusettings *usettings;

            if((usettings = find_user_settings(entry->curr->uid)) != NULL)
              if(usettings->exempt & INSTANT_PURGE_DT)
              {
                entry->curr->daily_time_used = 0;
                entry->curr->no_dtime_warnings = 0;
                sprintf(str,"N: '%s' Daily Time instantly reset",
                  entry->curr->user);
                log_msg(str,0);
              }
          }
          break;
        }
      if(entry->next != NULL) x = set_actives_gotoend(entry->next);
    }
    else
      x = entry;
    return x;
  }

  void get_user_entries(tutmp_ll *entry)
  {
    register int exit_rl = 1;

    do
    {
      if(fread(&tmp_buf,sizeof(struct utmp),1,fp) == 1)
      {
        register int x;
        char user_name[STR_USER_LEN];

        for(x = 0;x < (STR_USER_LEN - 1);x++)
          user_name[x] = tmp_buf.ut_user[x];
        user_name[x] = 0;
        if((tmp_buf.ut_type == USER_PROCESS) &&
           (find_user(&utmp_ll_root,user_name) == NULL))
        {
          tusettings *usettings;
          struct passwd *pw_entry = getpwnam(user_name);
          if(pw_entry != NULL)
            if(pw_entry->pw_uid)
              if((usettings = find_user_settings(pw_entry->pw_uid)) != NULL)
              {
                exit_rl = 0;
                if((entry->curr = malloc(sizeof(tutmp))) != NULL)
                {
                  strcpy(entry->curr->user,user_name);
                  entry->curr->uid = pw_entry->pw_uid;
                  entry->curr->pid = tmp_buf.ut_pid;
                  entry->curr->active = 1;
                  entry->curr->total_time_used =
                    (timer - tmp_buf.ut_time) - (timer - old_timer);
                  entry->curr->daily_time_used = entry->curr->total_time_used;
                  entry->curr->no_ttime_warnings = 0;
                  entry->curr->no_dtime_warnings = 0;
                  entry->curr->old_jiffies = 0;
                  entry->curr->jiffies = 0;
                  entry->curr->rss = 0;
                  entry->curr->no_cpu_warnings = 0;
                  entry->curr->no_mem_warnings = 0;
                  entry->curr->nice_factor = usettings->nice_factor;
                  strncpy(substr,ctime(&(tmp_buf.ut_time)) + 11,5);
                  substr[5] = 0;
                  sprintf(str,"Logged: New user '%s' at %s",
                    entry->curr->user,substr);
                  log_msg(str,1);

                  entry->next = malloc(sizeof(tutmp_ll));
                  if(entry->next != NULL)
                  {
                    entry->next->curr = NULL;
                    entry->next->next = NULL;
                    get_user_entries(entry->next);
                  }
                  else
                    log_msg("E: Couldn't malloc 'entry->next' (get_user_entries)",1);
                }
                else
                  log_msg("E: Couldn't malloc 'entry->curr' (get_user_entries)",1);
              }
        }
      }
      else
        exit_rl = 0;
    } while(exit_rl);
  }
  tutmp_ll *update_entry = NULL;

  if((fp = fopen(UTMP_FILE,"rb")) == NULL)
  {
    log_msg("E: Couldn't open " UTMP_FILE " (get_active_users)",1);
    return 0;
  }
  update_entry = set_actives_gotoend(&utmp_ll_root);
  rewind(fp);
  get_user_entries(update_entry);
  fclose(fp);
  return 1;
}

void free_user_entries(tutmp_ll *entry)
{
  if(entry->next != NULL) free_user_entries(entry->next);
  free(entry->next);
  if(entry->curr != NULL) free(entry->curr);
  entry->next = NULL;
  entry->curr = NULL;
}

/***************************************************************************
* load users (tutmp_ll) doc/clobberd.db
***************************************************************************/
void load_users(void)
{
  char str[STR_LEN];
  FILE *fp;

  void load_user(tutmp_ll *entry)
  {
    entry->curr = malloc(sizeof(tutmp));
    if(entry->curr != NULL)
    {
      if(fread(entry->curr,sizeof(tutmp),1,fp) == 1)
      {
        sprintf(str,"I: Recovering '%s' from '" DATA_FILE "'",entry->curr->user);
        log_msg(str,0);
        entry->next = malloc(sizeof(tutmp_ll));
        if(entry->next != NULL)
        {
          entry->next->curr = NULL;
          entry->next->next = NULL;
          load_user(entry->next);
        }
        else
          log_msg("E: Couldn't malloc 'entry->next' (load_user)",1);
      }
      else
      {
        free(entry->curr);
        entry->curr = NULL;
      }
    }
    else
      log_msg("E: Couldn't malloc 'entry->curr' (load_user)",1);
  }

  if((fp = fopen(DATA_FILE,"rb")) == NULL)
  {
    log_msg("W: Couldn't open '" DATA_FILE "' (load_users). This is "
            "no big deal if first installed",1);
    return;
  }
  if(fread(&utmp_hr,1,sizeof(tutmp_hr),fp) == sizeof(utmp_hr))
    load_user(&utmp_ll_root);
  fclose(fp);
}

/***************************************************************************
* save users (tutmp_ll) doc/clobberd.db
***************************************************************************/
void save_users(void)
{
  FILE *fp;
  int reload = 0;
  char str[STR_LEN];

  void save_user(tutmp_ll *entry)
  {
    if(entry->curr != NULL)
    {
      if(getpwuid(entry->curr->uid) != NULL)
      {
        if(fwrite(entry->curr,sizeof(tutmp),1,fp) == sizeof(tutmp))
        {
          log_msg("E: writing '"DATA_FILE"' (save_user).. pretty serious shit."
                  " Will forget user tracking if daemon dies",1);
          return;
        }
      }
      else
      {
        sprintf(str,"I: Ditching '%s' in '%s', as no longer a user"
                    " in '/etc/passwd'",entry->curr->user,DATA_FILE);
        log_msg(str,1);
        reload = 1;
      }
      if(entry->next != NULL) save_user(entry->next);
    }
  }

  if((fp = fopen(DATA_FILE,"wb")) == NULL)
  {
    log_msg("E: Couldn't save '" DATA_FILE "' (save_users)",1);
    return;
  }
  if(fwrite(&utmp_hr,1,sizeof(tutmp_hr),fp) == sizeof(utmp_hr))
    save_user(&utmp_ll_root);
  fclose(fp);
  if(reload)
  {
    sprintf(str,"I: Reloading '%s' due to purge",DATA_FILE);
    log_msg(str,1);
    free_user_entries(&utmp_ll_root);
    load_users();
  }
}

/***************************************************************************
* Midnight reset. purges all warnings, and daily time usage.
***************************************************************************/
void reset_at_midnight(void)
{
  struct tm *now;

  void reset_daily_stuff(tutmp_ll *entry)
  {
    if(entry->curr != NULL)
    {
      entry->curr->daily_time_used = 0;
      entry->curr->no_dtime_warnings = 0;
      entry->curr->no_cpu_warnings = 0;
      entry->curr->no_mem_warnings = 0;
      if(entry->next != NULL) reset_daily_stuff(entry->next);
    }
  }
  now = localtime(&timer);
  if((midnight_reset_f) || (utmp_hr.yesterday != now->tm_mday))
  {
    utmp_hr.yesterday = now->tm_mday;
    midnight_reset_f = 0;
    reset_daily_stuff(&utmp_ll_root);
    log_msg("Resetting: Midnight reset... and it's a brand new day",1);
  }
}

/***************************************************************************
* Total Purge
***************************************************************************/
void total_purge(void)
{
  if(total_purge_f)
  {
    total_purge_f = 0;
    free_user_entries(&utmp_ll_root);
    log_msg("Resetting: User Monitorings are now Totally P U R G E D.",1);
  }
}

/***************************************************************************
* warning stuff..
***************************************************************************/
#define tty_printf(s,d) write(s,d,strlen(d))

void warn_user(uid_t uid,char *email_str,char *message)
{
  char str[STR_LEN];
  struct utmp tmp_buf;
  FILE *fp;
  struct passwd *user;

  if((user = getpwuid(uid)) == NULL)
  {
    sprintf(str,"E: Unknown user uid %d (warn_user)",uid);
      log_msg(str,1);
    return;
  }
  if(email_str[0] == '*') /*do tty*/
  {
    if((fp = fopen(UTMP_FILE,"rb")) == NULL)
    {
      sprintf(str,
              "E: Couldn't open %s (warn_user). '%s' wasn't warned",
              UTMP_FILE,user->pw_name);
      log_msg(str,1);
      return;
    }
    while(fread(&tmp_buf,sizeof(struct utmp),1,fp) == 1)
    {
      if(strncmp(tmp_buf.ut_user,user->pw_name,UT_NAMESIZE) == 0)
      {
        register int fp2;
        char substr[STR_LEN];

        strcpy(substr,"/dev/");
        strcat(substr,tmp_buf.ut_line);
        if((fp2 = open(substr,O_WRONLY | O_NOCTTY | O_APPEND | O_NONBLOCK)) != -1)
        {
          tty_printf(fp2,"\a\n");
          tty_printf(fp2,"--\n");
          tty_printf(fp2,"From: " EMAIL_ADDR "\n");
          sprintf(str,"Mesg: %s\n",message);
          tty_printf(fp2,str);
          close(fp2);
        }
        else
        {
          sprintf(str,
                  "E: '%s' couldn't be opened. %s may not be warned",
                  substr,user->pw_name);
          log_msg(str,1);
        }
      }
    }
    fclose(fp);
  }
  else /*if email*/
  {
    char *s;
    register int socket_d;

    if(!email_str[0])
      s = user->pw_name;
    else
      s = email_str;
    if((socket_d = tcpip_client_creat(our_hostname,25)) != -1)
    {
      sprintf(str,
        "HELO %s\n"
        "MAIL FROM: %s\n"
        "RCPT TO: %s\n"
        "DATA\n"
        "SUBJECT: Warning\n\n"
        "%s\n\n"
        "--\n"
        "%s\n"
        "\n.\nQUIT\n",
        our_hostname,EMAIL_ADDR,s,message,SIGNATURE_STR);
      net_printf(socket_d,str);
      shutdown(socket_d,2);
    }
    else
    {
      sprintf(str,"E: Unknown host '%s'. '%s' couldn't be warned\n",
        our_hostname,s);
      log_msg(str,1);
    }
  }
}

/***************************************************************************
* By ?? (2.4D, Jan 97)
***************************************************************************/
int dialins_free(void)
{
  register int free_dialins = 0;
  char *newline;
  FILE *fp;
  char ttyname[STR_LEN];
  struct stat stbuf;

  if((fp = fopen(DIALINS_FILE, "r")) != NULL)
  {
    while(!feof(fp))
      if(fgets(ttyname, STR_LEN, fp) != NULL)
      {
        if((newline = strstr(ttyname, "\n")) != NULL)
          *newline = '\0';
        if(stat(ttyname, &stbuf) == 0)
          if(stbuf.st_uid == 0)
            free_dialins++;
      }
    fclose(fp);
  }
  else
    return -1;
  return free_dialins;
}

/***************************************************************************
* monitor user load..
*
* i have discovered that sweeping the /proc structure can become quite
* cpu intensive..
***************************************************************************/
void get_agg_stats(int *tot_cpu_jiffies,unsigned int *tot_rss)
{
  FILE *fp;
 
  if((fp = fopen("/proc/stat","r")) == NULL)
  {
    log_msg("E: Couldn't open '/proc/stat' (get_agg_stats)",1);
    return;
  }
  do
  {
    char str[STR_LEN];

    fgets(str,STR_LEN,fp);
    if(strstr(str,"cpu") != NULL)
    {
      int a,b,c,d;
 
      sscanf(str,"cpu %d %d %d %d",&a,&b,&c,&d);
      *tot_cpu_jiffies = a + b + c + d;
    }
    if(strstr(str,"page") != NULL)
      sscanf(str,"page %u %*u",tot_rss);
  } while(!feof(fp));
  fclose(fp);
}   

void get_user_loads(void)
{
  int l_jiffies;
  unsigned int l_rss;
  FILE *fp;
  DIR *dp;
  struct dirent *dir_entry;
  struct stat statbuf;
  char str[STR_LEN];

  void clear_loads(tutmp_ll *entry)
  {
    if(entry->curr != NULL)
    {
      entry->curr->old_jiffies = entry->curr->jiffies;
      entry->curr->jiffies = 0;
      entry->curr->rss = 0;
      if(entry->next != NULL) clear_loads(entry->next);
    }
  }
  void add_loads(tutmp_ll *entry)
  {
    if(entry->curr != NULL)
      if(entry->curr->uid == statbuf.st_uid)
      {
        entry->curr->jiffies += l_jiffies;
        entry->curr->rss += l_rss;
      }
      else
        if(entry->next != NULL) add_loads(entry->next);
  }

  old_total_jiffies = total_jiffies;
  get_agg_stats(&total_jiffies,&total_rss);
  clear_loads(&utmp_ll_root);
  if((dp = opendir("/proc")) == NULL)
  {
    log_msg("E: Couldn't open '/proc' (get_user_loads)",1);
    return;
  }
  while((dir_entry = readdir(dp)) != NULL)
  {
    int a;

    if(sscanf(dir_entry->d_name,"%d",&a))
    {
      strcpy(str,"/proc/");
      strcat(str,dir_entry->d_name);
      if(lstat(str,&statbuf) != -1)
      {
        strcpy(str,"/proc/");
        strcat(str,dir_entry->d_name);
        strcat(str,"/stat");
        if((fp = fopen(str,"r")) == NULL)
        {
          log_msg("E: Couldn't open '/proc/[pid]/stat' (get_user_loads)",1);
          return;
        }
        fscanf(fp,
          "%*d %*s %*c %*d %*d %*d %*d %*d %*u %*u %*u %*u %*u"
      /*utime stime                             rss*/
          "%d %d %*d %*d %*d %*d %*u %*u %*d %*u %u",
          &l_jiffies,&a,&l_rss);
        fclose(fp);
        l_jiffies += a;
        add_loads(&utmp_ll_root);
      }
    }
  }
  closedir(dp);
  no_dialins_free = dialins_free();
}

/***************************************************************************
* Clobber user...
***************************************************************************/
void clobber_user(tutmp_ll *entry)
{
  DIR *dp;
  struct dirent *dir_entry;
  struct stat statbuf;
  char str[STR_LEN];

  entry->curr->active = 0;
  killpg(entry->curr->pid,usettings_hr.clobber_sig);
  if((dp = opendir("/proc")) == NULL)
  {
    log_msg("E: Couldn't open '/proc' (clobber_user)",1);
    return;
  }
  while((dir_entry = readdir(dp)) != NULL)
  {
    int e_pid;

    if(sscanf(dir_entry->d_name,"%d",&e_pid))
    {
      strcpy(str,"/proc/");
      strcat(str,dir_entry->d_name);
      if((lstat(str,&statbuf) != -1) &&
         (entry->curr->uid == statbuf.st_uid))
        kill(e_pid,usettings_hr.clobber_sig);
    }
  }
  closedir(dp);
}

/***************************************************************************
* User monitoring routines
***************************************************************************/
time_t timer_used;

/**************************************
*
**************************************/
void abuse_clobber(tutmp_ll *entry,tusettings *us_entry)
{
  if(us_entry->exempt & EXEMPT_CLOBBERING)
  {
    char str[STR_LEN];

    entry->curr->nice_factor += 1.0;
    if((int)entry->curr->nice_factor > 19)
      entry->curr->nice_factor = 19;
    sprintf(str,"N: '%s' Real renice level increased to %f",
      entry->curr->user,entry->curr->nice_factor);
    log_msg(str,1);
  }
  else
    clobber_user(entry);
}

/**************************************
*
**************************************/
void monitor_total_time(tutmp_ll *entry,tusettings *us_entry)
{
  if((entry->curr->active) &&
     !(us_entry->exempt & EXEMPT_TTIME))
  {
    char str[STR_LEN];

    entry->curr->total_time_used += timer_used;
    sprintf(str,"M: Total time for '%s': %ds",
      entry->curr->user,(int)entry->curr->total_time_used);
    log_msg(str,0);
    if(entry->curr->total_time_used > us_entry->total_time_limit)
      if(entry->curr->no_ttime_warnings >= usettings_hr.no_of_warnings)
      {
        warn_user(entry->curr->uid,us_entry->email,TTIME_MESS);
        clobber_user(entry);
        sprintf(str,"N: '%s' clobberd.. Total time limit expired",
          entry->curr->user);
        log_msg(str,1);
      }
      else
      {
        entry->curr->no_ttime_warnings++;
        sprintf(str,TTIME_WARN,entry->curr->no_ttime_warnings,
          usettings_hr.no_of_warnings);
        warn_user(entry->curr->uid,us_entry->email,str);
        sprintf(
          str,
          "N: '%s' Total time limit expired. Warning No. %d/%d",
          entry->curr->user,entry->curr->no_ttime_warnings,
          usettings_hr.no_of_warnings);
        log_msg(str,1);
      }
  }
}

/**************************************
*
**************************************/
void monitor_daily_time(tutmp_ll *entry,tusettings *us_entry)
{
  if((entry->curr->active) && !(us_entry->exempt & EXEMPT_DTIME))
  {
    char str[STR_LEN];

    if((no_dialins_free != -1) &&
       (no_dialins_free >= usettings_hr.min_free_dialins))
    {
      sprintf(str,
        "M: Low usage (Dialin Stat: %d/%d)."
        " '%s' is exempt from daily time monitoring",
        no_dialins_free,usettings_hr.min_free_dialins,entry->curr->user);
      log_msg(str,0);
    }
    else
    {
      entry->curr->daily_time_used += timer_used;
      sprintf(str,"M: Daily time for '%s': %ds",
        entry->curr->user,(int)entry->curr->daily_time_used);
      log_msg(str,0);
      if(entry->curr->daily_time_used > us_entry->daily_time_limit)
        if(entry->curr->no_dtime_warnings >= usettings_hr.no_of_warnings)
        {
          warn_user(entry->curr->uid,us_entry->email,DTIME_MESS);
          clobber_user(entry);
          sprintf(str,"N: '%s' clobberd.. Daily time limit expired",
            entry->curr->user);
          log_msg(str,1);
        }
        else
        {
          entry->curr->no_dtime_warnings++;
          sprintf(str,DTIME_WARN,entry->curr->no_dtime_warnings,
            usettings_hr.no_of_warnings);
          warn_user(entry->curr->uid,us_entry->email,str);
          sprintf(
            str,
            "N: '%s' Daily time limit expired. Warning No. %d/%d",
            entry->curr->user,entry->curr->no_dtime_warnings,
            usettings_hr.no_of_warnings);
            log_msg(str,1);
        }
    }
  }
}

/**************************************
*
**************************************/
void monitor_cpu_load(tutmp_ll *entry,tusettings *us_entry)
{
  if((entry->curr->active) &&
     !(us_entry->exempt & EXEMPT_CPU))
  {
    char str[STR_LEN];
    register int cpu_usage =
      ((entry->curr->jiffies - entry->curr->old_jiffies) * 100) /
      (total_jiffies - old_total_jiffies);
    if(cpu_usage < 0) cpu_usage = 0;
    sprintf(str,"M: CPU usage for '%s' is at %d%%",
      entry->curr->user,cpu_usage);
    log_msg(str,0);
    if(cpu_usage > us_entry->cpu_limit)
      if(entry->curr->no_cpu_warnings >= usettings_hr.no_of_warnings)
      {
        sprintf(str,CPU_MESS,cpu_usage);
        warn_user(entry->curr->uid,us_entry->email,str);
        abuse_clobber(entry,us_entry);
        sprintf(str,"N: '%s' clobberd.. CPU abuse (CPU usage: %d%%)",
          entry->curr->user,cpu_usage);
        log_msg(str,1);
      }
      else
      {
        entry->curr->no_cpu_warnings++;
        sprintf(str,CPU_WARN,entry->curr->no_cpu_warnings,
          usettings_hr.no_of_warnings,cpu_usage);
        warn_user(entry->curr->uid,us_entry->email,str);
        sprintf(
          str,
          "N: '%s' pounding CPU. Warning No. %d/%d (CPU usage: %d%%)",
          entry->curr->user,entry->curr->no_cpu_warnings,
          usettings_hr.no_of_warnings,cpu_usage);
          log_msg(str,1);
      }
  }
}

/**************************************
*
**************************************/
void monitor_mem_load(tutmp_ll *entry,tusettings *us_entry)
{
  if((entry->curr->active) &&
     !(us_entry->exempt & EXEMPT_MEM))
  {
    char str[STR_LEN];
    register int mem_usage = (entry->curr->rss * 100) / total_rss;
    if(mem_usage < 0) mem_usage = 0;
    sprintf(str,"M: Memory usage for '%s' is at %d%%",
      entry->curr->user,mem_usage);
    log_msg(str,0);
    if(mem_usage > us_entry->mem_limit)
      if(entry->curr->no_mem_warnings >= usettings_hr.no_of_warnings)
      {
        sprintf(str,MEM_MESS,mem_usage);
        warn_user(entry->curr->uid,us_entry->email,str);
        abuse_clobber(entry,us_entry);
        sprintf(str,"N: '%s' clobberd.. Memory abuse (Memory usage: %d%%)",
          entry->curr->user,mem_usage);
        log_msg(str,1);
      }
      else
      {
        entry->curr->no_mem_warnings++;
        sprintf(str,MEM_WARN,entry->curr->no_mem_warnings,
          usettings_hr.no_of_warnings,mem_usage);
        warn_user(entry->curr->uid,us_entry->email,str);
        sprintf(
          str,
          "N: '%s' pounding Memory. Warning No. %d/%d (Memory usage: %d%%)",
          entry->curr->user,entry->curr->no_mem_warnings,
          usettings_hr.no_of_warnings,mem_usage);
          log_msg(str,1);
      }
  }
}

/**************************************
*
**************************************/
void monitor_nice(tutmp_ll *entry,tusettings *us_entry)
{
  if((entry->curr->active) &&
     !(us_entry->exempt & EXEMPT_NICE))
  {
    float ndiff;
    char str[STR_LEN];

    if((int)entry->curr->nice_factor !=
      getpriority(PRIO_USER,entry->curr->uid))
    {
      setpriority(PRIO_USER,entry->curr->uid,
        (int)entry->curr->nice_factor);
      sprintf(str,"M: Set '%s' to nice %d",
        entry->curr->user,(int)entry->curr->nice_factor);
      log_msg(str,1);
    }
    ndiff = entry->curr->nice_factor - (float)us_entry->nice_factor;
    if(ndiff != 0)
    {
      if(ndiff > 0.25) ndiff = 0.25;
      if(ndiff < -0.25) ndiff = -0.25;
      entry->curr->nice_factor -= ndiff;
      sprintf(str,"M: '%s' Real niceness recovered to %f (Target: %d)",
        entry->curr->user,entry->curr->nice_factor,us_entry->nice_factor);
      log_msg(str,0);
    }
  }
}

/**************************************
*
**************************************/
void monitor_users(tutmp_ll *entry)
{
  char str[STR_LEN];

  void monitor_user(tutmp_ll *entry)
  {
    if(entry->curr != NULL)
    {
      if(entry->curr->active)
      {
        tusettings *us_entry;

        if((us_entry = find_user_settings(entry->curr->uid)) != NULL)
        {
          monitor_total_time(entry,us_entry);
          monitor_daily_time(entry,us_entry);
          monitor_cpu_load(entry,us_entry);
          monitor_mem_load(entry,us_entry);
          monitor_nice(entry,us_entry);
        }
        else
        {
          sprintf(str,"M: '%s' is not marked. User ignored",entry->curr->user);
          log_msg(str,0);
        }
      }
      else
      {
        sprintf(str,"M: '%s' is not logged. User ignored",entry->curr->user);
        log_msg(str,0);
      }
    }
    if(entry->next != NULL) monitor_user(entry->next);
  }

  timer_used = timer - old_timer;
  monitor_user(entry);
}

/***************************************************************************
*
***************************************************************************/
void print_user_info(int K)
{
  char str[STR_LEN],substr[STR_LEN];

  void print_user_entries(tutmp_ll *entry)
  {
    if(entry->curr != NULL)
    {
      tusettings *us_entry;

      if((us_entry = find_user_settings(entry->curr->uid)) != NULL)
      {
        strcpy(str,
        "         Yes                                                     Yes Yes\n");

        memcpy(str + 0,entry->curr->user,strlen(entry->curr->user));
        if(!entry->curr->active) memcpy(str + 9,"No ",3);

        if(us_entry->exempt & EXEMPT_TTIME)
          strcpy(substr,"*");
        else
          str_hh_mm(us_entry->total_time_limit - entry->curr->total_time_used,substr);
        memcpy(str + 13,substr,strlen(substr));

        if(us_entry->exempt & EXEMPT_DTIME)
          strcpy(substr,"*");
        else
          str_hh_mm(us_entry->daily_time_limit - entry->curr->daily_time_used,substr);
        memcpy(str + 21,substr,strlen(substr));

        if(us_entry->exempt & EXEMPT_NICE)
          strcpy(substr,"*");
        else
          sprintf(substr,"%d",(int)entry->curr->nice_factor);
        memcpy(str + 29,substr,strlen(substr));

        if(us_entry->exempt & EXEMPT_CPU)
          strcpy(substr,"*");
        else
        {
          register int cpu_usage =
            ((entry->curr->jiffies - entry->curr->old_jiffies) * 100) /
            (total_jiffies - old_total_jiffies);
          if(cpu_usage < 0) cpu_usage = 0;
          sprintf(substr,"%d/%d",cpu_usage,us_entry->cpu_limit);
        }
        memcpy(str + 33,substr,strlen(substr));

        if(us_entry->exempt & EXEMPT_MEM)
          strcpy(substr,"*");
        else
        {
          register int mem_usage = (entry->curr->rss * 100) / total_rss;
          if(mem_usage < 0) mem_usage = 0;
          sprintf(substr,"%d/%d",mem_usage,us_entry->mem_limit);
        }
        memcpy(str + 41,substr,strlen(substr));

        if(us_entry->exempt & EXEMPT_TTIME)
          strcpy(substr,"*");
        else
          sprintf(substr,"%d",entry->curr->no_ttime_warnings);
        memcpy(str + 50,substr,strlen(substr));

        if(us_entry->exempt & EXEMPT_DTIME)
          strcpy(substr,"*");
        else
          sprintf(substr,"%d",entry->curr->no_dtime_warnings);
        memcpy(str + 55,substr,strlen(substr));

        if(us_entry->exempt & EXEMPT_CPU)
          strcpy(substr,"*");
        else
          sprintf(substr,"%d",entry->curr->no_cpu_warnings);
        memcpy(str + 58,substr,strlen(substr));

        if(us_entry->exempt & EXEMPT_MEM)
          strcpy(substr,"*");
        else
          sprintf(substr,"%d",entry->curr->no_mem_warnings);
        memcpy(str + 62,substr,strlen(substr));

        if(!(us_entry->exempt & EXEMPT_CLOBBERING)) memcpy(str + 65,"No ",3);

        if(!(us_entry->exempt & INSTANT_PURGE_DT)) memcpy(str + 69,"No ",3);

        net_printf(K,str);
      }
      if(entry->next != NULL) print_user_entries(entry->next);
    }
  }

  sprintf(str,"\nClobberd-%s@%s - %s\n\n",
    VERSION,our_hostname,PLUG_TEXT);
  net_printf(K,str);
  sprintf(str,"      Run Level: %d\n",usettings_hr.run_level);
  net_printf(K,str);
  sprintf(str,"     Sleep Time: %ds\n",usettings_hr.sleep_time);
  net_printf(K,str);
  sprintf(str,"  Warning Level: %d\n",usettings_hr.no_of_warnings);
  net_printf(K,str);
  sprintf(str," Clobber Signal: %hd\n",usettings_hr.clobber_sig);
  net_printf(K,str);
  strcpy(str,"Min Free Modems: ");
  if(no_dialins_free == -1)
    strcat(str,"Inactive");
  else
  {
    sprintf(substr,"%hd/%hd",no_dialins_free,usettings_hr.min_free_dialins);
    strcat(str,substr);
    if(no_dialins_free >= usettings_hr.min_free_dialins)
      strcat(str," (low usage)");
  }
  strcat(str,"\n");
  net_printf(K,str);
  sprintf(str,"Default Setings: %s\n\n",
    usettings_hr.def_usettings_enab ? "Yes" : "No");
  net_printf(K,str);
  if(edit_mode != MONITOR)
    net_printf(K,"Admin maintenance. Clobberd suspended.. Come back later.\n\n");
  else
  {
    net_printf(K,"         Act Total   Daily        %      %        No of Warnings Cbr Ins\n");
    net_printf(K,"Name     ive TimeL   TimeL   Nce CPU/Lim Mem/Lim  TT  DT  Cu  Mm Abs DTR\n");
    net_printf(K,"========================================================================\n");
    if(utmp_ll_root.curr != NULL)
      print_user_entries(&utmp_ll_root);
    else
      net_printf(K,"\nNo users.\n");
    net_printf(K,"\n");
  }
}

/***************************************************************************
*
***************************************************************************/
void talk_to_telnets(void)
{
  time_t ltime = time(NULL);
  log_msg("I: -------Daemon Netting-------",0);
  listen_loop = 1;
  while(listen_loop)
  {
    struct timeval tv = {0,0};
    register int x;

    for(x = 0;x < MAX_CONNECTIONS;x++)
    {
      char str[STR_LEN];
      fd_set fds;

      FD_ZERO(&fds);
      FD_SET(0,&fds);
      FD_SET(main_socket,&fds);
      select(main_socket + 1,&fds,NULL,NULL,&tv);
      if(FD_ISSET(main_socket,&fds))
      {
        struct sockaddr_in sin;
        register int s;

        s = accept(main_socket,(struct sockaddr *)&sin,&newaddrsize);
        if(s != -1)
        {
          struct hostent *newhost;

          fcntl(s,F_SETFL,O_NONBLOCK);
          newhost = gethostbyaddr((char *)&(sin.sin_addr.s_addr),4,AF_INET);
          if(newhost != NULL)
            sprintf(str,"K: Telnet Connection from [%s]",newhost->h_name);
          else
            sprintf(str,"K: Telnet Connection from [UNKNOWN]");
          log_msg(str,1);
          print_user_info(s);
          shutdown(s,2);
          close(s);
        }
      }
    }
    sleep(1);
    if((time(NULL) - ltime) > usettings_hr.sleep_time) break;
  }
}

/***************************************************************************
*
***************************************************************************/
void prime_loop(void)
{
  timer = time(NULL);
  old_timer = timer;
  get_agg_stats(&total_jiffies,&total_rss);
  old_total_jiffies = 0;
  if(get_user_settings())
  {  
    load_users();
    if(get_active_users()) get_user_loads();
  }
}

void main_loop(void)
{
  while(running)
  {
    if(edit_mode != EDIT)
    {
      log_msg("I: -------Daemon Monitoring----",0);
      old_timer = timer;
      timer = time(NULL);
      if(edit_mode == RELOAD)
      {
        log_msg("I: Reloading " CONFIG_FILE "...",1);
        free_user_settings();
        if(get_user_settings())
        {
          log_msg("I: Reloading " DATA_FILE "...",1);
          free_user_entries(&utmp_ll_root);
          load_users();
          edit_mode = MONITOR;
        }
      }
      if(get_active_users())
      {
        reset_at_midnight();
        total_purge();
        get_user_loads();
        monitor_users(&utmp_ll_root);
      }
      save_users();
    }
    else
      log_msg("I: -------Daemon in Edit mode--",0);
    talk_to_telnets();
  }
}

/***************************************************************************
*
***************************************************************************/
int main(int argc,char *argv[])
{
  register int x;
  char str[STR_LEN];

  printf(WELCOME);
  for(x = 0;x < argc;x++)
  {
    if(argv[x][0] == '-')
    {
      switch(argv[x][1])
      {
        case 'h':
          printf(
            "\nUsage: clobberd <options>\n"
             " -h        Display this help info\n"
             " -!        Run as normal process, logs sent to screen.\n"
             "\n"
          );
          return 1;
        case '!':
          daemon_mode = 0;
          break;
      }
    }
  }
  sprintf(str,"(%s)",argv[0]);
  if(find_process(str,0) > 0)
  {
    printf("Turkey, You are already running clobberd. Aborting this process..\a\n");
    return -1;
  }
  if(gethostname(our_hostname,STR_LEN) != -1)
  {
    init();
    daemonise();
    if((main_socket = tcpip_serv_creat()) != -1)
    {
      log_msg("-----Clobberd Fired Up-----",1);
      sprintf(str,"I: On %s:%d and priming..",our_hostname,NET_PORT);
      log_msg(str,1);
      prime_loop();
      main_loop();
      free_user_settings();
      free_user_entries(&utmp_ll_root);
      log_msg("-----Clobberd End----------",1);
      tcpip_serv_reaper();
    }
  }
  else
    printf("Error: Couldn't resolve current hostname. Aborting\n");
  return 1;
}
