/***************************************************************************
****************************************************************************
****************************************************************************
*
* Clobberd - By Jason Nunn
* FREEWARE. Authorship Reserved 1996
*
* "A user regulator with attitude.." (ANSI C)
*
* Snail: 32 Rothdale Road, Moil, Darwin, NT, 0810, Australia
*
* ==================================================================
*
* Developed on a Linux 1.2.13 box ...then later 2.0.0
*
****************************************************************************
****************************************************************************
***************************************************************************/
#include CURSES_LOC
#include <unistd.h>
#include <malloc.h>
#include <utmp.h>
#include <pwd.h>
#include <sys/types.h>
#include <dirent.h>
#include <signal.h>
#include "clobberd.h"
#include "misc.h"
#define S_STR_LEN 16
#define USETTINGS_SCROLL_MAX 4

int ch = 0,endedit = 0,editmode = 0;

tusettings_hr usettings_hr = DEFAULT_HR;
int           pw_table_size = 0;
int           usettings_list_size = 0;
tusettings    *usettings_list = NULL;
tutmp_hr      utmp_hr;
int           utmp_list_size = 0;
tutmp         *utmp_list = NULL;

int           def_hlpos = 0;

int           usettings_option = 0;
int           usettings_pos_real = 0;
int           usettings_pos_hl  = 0;

int           utmp_option = 0;
int           utmp_pos_real = 0;
int           utmp_pos_hl  = 0;

/***************************************************************************
*
***************************************************************************/
int load_pw_table(void)
{
  setpwent();
  while(getpwent() != NULL) pw_table_size++;
  endpwent();
  return pw_table_size;
}

/***************************************************************************
*
***************************************************************************/
int load_usettings(void)
{
  FILE *fp;
  register int c;

  if((usettings_list = malloc(sizeof(tusettings) *
                              pw_table_size)) == NULL)
  {
    printf("Couldn't Alloc 'usettings_list' (load_usettings).\n");
    return 0;
  }
  if((fp = fopen(CONFIG_FILE,"rb")) == NULL)
  {
    printf("Error: Couldn't open " CONFIG_FILE " (load_usettings).\n");
    return 1;
  }
  if(fread(&usettings_hr,1,sizeof(tusettings_hr),fp) == sizeof(usettings_hr))
    for(c = 0;c < pw_table_size;c++)
    {
      if(fread((usettings_list + usettings_list_size),1,
        sizeof(tusettings),fp) == sizeof(tusettings))
      {
        if(usettings_list_size++ >= pw_table_size)
        {
          printf("Warning: " CONFIG_FILE " entries exceed users (load_usettings).\n");
          break;
        }
      }
      else
        break;
    }
  fclose(fp);
  return 1;
}

/***************************************************************************
*
***************************************************************************/
int load_utmp(void)
{
  FILE *fp;
  register int c;

  if((utmp_list = malloc(sizeof(tutmp) * pw_table_size)) == NULL)
  {
    printf("Couldn't Alloc 'utmp_list' (load_utmp)\n");
    return 0;
  }
  if((fp = fopen(DATA_FILE,"rb")) == NULL)
  {
    printf("Error: Couldn't open " DATA_FILE " (load_utmp)\n");
    return 1;
  }
  if(fread(&utmp_hr,1,sizeof(tutmp_hr),fp) == sizeof(utmp_hr))
    for(c = 0;c < pw_table_size;c++)
    {
      if(fread((utmp_list + utmp_list_size),1,sizeof(tutmp),fp) == sizeof(tutmp))
      {
        if(utmp_list_size++ >= pw_table_size)
        {
          printf("Warning: " DATA_FILE " entries exceed users (load_utmp).\n");
          break;
        }
      }
      else
        break;
    }
  fclose(fp);
  return 1;
}

/***************************************************************************
*
***************************************************************************/
void display_topbar(void)
{
  char s[81];

  clear();
  move(0,0);
  attron(A_REVERSE);
  sprintf(s," Clobberd %-4s%66s", VERSION, "by Jason Nunn ");
  addstr(s);
}

void hl_bar(char option,char x)
{
  if(option == x)
    attron(A_REVERSE);
  else
    attroff(A_REVERSE);
}

void update_screen(void)
{
  attroff(A_REVERSE);
  move(0,0);
  refresh();
}

void display_mm_ss(int t)
{
  char ms[4],s[8];

  if(t < 0) t = 0;
  sprintf(ms,"%d",(int)(t % 60));
  if(strlen(ms) == 1)
  {
    ms[1] = ms[0];
    ms[0] = '0';
    ms[2] = 0;
  }
  sprintf(s,"%d:%s",(int)(t / 60),ms);
  addstr(s);
}

void range_check(int *val,int lim1,int lim2)
{
  if(*val < lim1) *val = lim1;
  if(*val > lim2) *val = lim2;
}

void display_hh_mm(int t)
{
  char s[S_STR_LEN];
  str_hh_mm(t,s);
  addstr(s);
}

void display_int(int x)
{
  char s[S_STR_LEN];
    
  sprintf(s,"%3d",x);
  addstr(s);
}

void display_sint(int x)
{  
  char s[S_STR_LEN];
  
  sprintf(s,"%d",x);
  addstr(s);
}

void display_real(float x)
{
  char s[S_STR_LEN];

  sprintf(s,"%10f",x);
  addstr(s);
}

void clear_area(char y1,char x1,char y2,char x2)
{
  register char a;
  for(y1 = y1;y1 <= y2;y1++)
    for(a = x1;a <= x2;a++)
    {
      move(y1,a);
      addch(' ');
    }
}

void get_string(char y, char x,char *str,int maxlen)
{
  register int a;

  attron(A_REVERSE);
  for(a = 0;a < maxlen;a++)
  {
    move(y,x + a);
    addch(' ');
  }
  echo();
  move(y,x);
  wgetnstr(stdscr,str,maxlen);
  noecho();
  attroff(A_REVERSE);
  clear_area(y,x,y,x + maxlen - 1);
}

int get_mm_ss(int y,int x,int strlen)
{
  char str[S_STR_LEN];
  int m = 0,s = 0;

  get_string(y,x,str,strlen);
  sscanf(str,"%d:%d",&m,&s);
  if(s > 59) s = 59;
  return s + (m * 60);
}

int get_hh_mm(int y,int x,int strlen)
{
  char str[S_STR_LEN];
  int h = 0,m = 0;

  get_string(y,x,str,strlen);
  sscanf(str,"%d:%d",&h,&m);
  if(m > 59) m = 59;
  return (m + (h * 60)) * 60;
}

int get_int(int y,int x,int strlen)
{
  int v = 0;
  char str[S_STR_LEN];

  get_string(y,x,str,strlen);
  sscanf(str,"%d",&v);
  return v;
}

float get_real(int y,int x,int strlen)
{
  float v = 0;
  char str[S_STR_LEN];

  get_string(y,x,str,strlen);
  sscanf(str,"%f",&v);
  return v;
}

void ferr_message(char *mess_str)
{
  attron(A_BLINK);
  move(23,0);
  addstr(mess_str);
  addstr(" [hit key]");
  attroff(A_BLINK);
  beep();
  update_screen();
  getch();
  clear_area(23,0,23,79);
  update_screen();
}

void cursor_up(int *pos_real,int *pos_hl,void (*scroll_routine)(void))
{
  if(*pos_hl > 0)
  {
    (*pos_hl)--;
    scroll_routine();
    update_screen();
  }
  else
  {
    if(*pos_real > 0) (*pos_real)--;
    scroll_routine();
    update_screen();
  }
}

void cursor_down(int *pos_real,
                 int *pos_hl,
                 int scroll_len,
                 int table_size,
                 void (*scroll_routine)(void))
{
  table_size--;
  scroll_len--;
  if(*pos_hl < scroll_len)
  {
    if(*pos_hl < table_size)
    {
      (*pos_hl)++;
      scroll_routine();
      update_screen();
    }
  }
  else
  {
    if(*pos_real < (table_size - scroll_len)) (*pos_real)++;
    scroll_routine();
    update_screen();
  }
}

/***************************************************************************
*
***************************************************************************/
void delete_entry(void *table,int tuple_size,int *entry_real,int *entry_hl,
                  int *table_size,void (*scroll_routine)(void))
{
  register int x,entry_no = *entry_real + *entry_hl;

  if((*table_size >= 0) && (entry_no < *table_size))
  {
    for(x = entry_no;x < *table_size;x++)
      memcpy(table + (x * tuple_size),
             table + ((x + 1) * tuple_size),tuple_size);
    (*table_size)--;
    scroll_routine();
    update_screen();
  }
}

/***************************************************************************
*
***************************************************************************/
int quit_confirm(void)
{
  attron(A_BLINK);
  clear_area(23,0,23,79);
  move(23,0);
  addstr("Quit yes? [y/N]");
  attroff(A_BLINK);
  beep();
  update_screen();
  ch = getch();
  clear_area(23,0,23,79);
  update_screen();
  if((ch == 'y') || (ch == 'Y'))
    return 1;
  ch = 0;
  return 0;
}

/***************************************************************************
*
***************************************************************************/
void save_files(void)
{
  register int c;
  FILE *fp;
  char *errmess = "Error: Writing " CONFIG_FILE " (save_files).";

  if((fp = fopen(CONFIG_FILE,"wb")) != NULL)
  {
    if(fwrite(&usettings_hr,1,sizeof(tusettings_hr),fp) ==
      sizeof(usettings_hr))
    {
      for(c = 0;c < usettings_list_size;c++)
        if((usettings_list + c)->uid != 65535)
          if(fwrite((usettings_list + c),1,sizeof(tusettings),fp) !=
            sizeof(tusettings))
          {
            ferr_message(errmess);
            break;
          }
    }
    else
      ferr_message(errmess);
    fclose(fp);
  }
  else
    ferr_message("Error: Couldn't open " CONFIG_FILE " (save_files)");

  if((fp = fopen(DATA_FILE,"wb")) != NULL)
  {
    register int blocksize = sizeof(tutmp) * utmp_list_size;
    if(fwrite(&utmp_hr,1,sizeof(tutmp_hr),fp) == sizeof(utmp_hr))
      if(fwrite(utmp_list,1,blocksize,fp) != blocksize)
        ferr_message("Error: Writing " DATA_FILE " (save_files).");
  }
  else
    ferr_message("Error: Couldn't open " DATA_FILE " (save_files)");
  fclose(fp);

  attroff(A_BLINK);
  move(23,0);
  addstr("Saved files.");
  update_screen();
}

/***************************************************************************
*
***************************************************************************/
void display_hr_titles(void)
{
  hl_bar(0,editmode);
  move(2,1);
  addstr("Run level.......:");

  hl_bar(1,editmode);
  move(3,1);
  addstr("Sleep time......:");

  hl_bar(2,editmode);
  move(4,1);
  addstr("No of warnings..:");

  hl_bar(3,editmode);
  move(5,1);
  addstr("Enable a Default:");

  hl_bar(4,editmode);
  move(2,30);
  addstr("Clobber Signal..:");

  hl_bar(5,editmode);
  move(3,30);
  addstr("Min Free Dialins:");

  hl_bar(6,editmode);
  move(4,30);
  addstr("Default Settings:");

  hl_bar(7,editmode);
  move(7,1);
  addstr("Settings:");

  hl_bar(8,editmode);
  move(15,1);
  addstr("Stats:");
}

/***************************************************************************
*
***************************************************************************/
void tt_field(tusettings *entry,int dis_pos)
{
  if(entry->exempt & EXEMPT_TTIME)
  {
    move(dis_pos,50);
    addstr("=E=");
  }   
  else
  {
    char str[STR_LEN];

    str_hh_mm(entry->total_time_limit,str);
    move(dis_pos,48 + (6 - strlen(str)));
    addstr(str);
  }
}

void dt_field(tusettings *entry,int dis_pos)
{
  if(entry->exempt & EXEMPT_DTIME)
  {
    move(dis_pos,56);
    addstr("=E=");
  }
  else
  {
    char str[STR_LEN];

    str_hh_mm(entry->daily_time_limit,str);
    move(dis_pos,55 + (5 - strlen(str)));
    addstr(str);
  }
}

void nice_field(tusettings *entry,int dis_pos)
{
  move(dis_pos,61);
  if(entry->exempt & EXEMPT_NICE)
    addstr("=E=");
  else
    display_int(entry->nice_factor);
}

void cpu_field(tusettings *entry,int dis_pos)
{
  move(dis_pos,65);
  if(entry->exempt & EXEMPT_CPU)
    addstr("=E=");
  else
    display_int(entry->cpu_limit);
}

void mem_field(tusettings *entry,int dis_pos)
{
  move(dis_pos,69);
  if(entry->exempt & EXEMPT_MEM)
    addstr("=E=");
  else
    display_int(entry->mem_limit);
}

void coa_field(tusettings *entry,int dis_pos)
{
  move(dis_pos,73);
  if(entry->exempt & EXEMPT_CLOBBERING)
    addch('y');
  else
    addch('n');
}

void iptd_field(tusettings *entry,int dis_pos)
{
  move(dis_pos,75);
  if(entry->exempt & INSTANT_PURGE_DT)
    addch('y');
  else
    addch('n');
}

/***************************************************************************
*  
***************************************************************************/
char *tit_bar1 = "  Total Daily Nic CPU Mem A I ";

void def_usettings_hl_bar(unsigned char x)
{
  if(def_hlpos == x)
    attron(A_REVERSE);
  else
    attroff(A_REVERSE);
}

void display_hr_vals(void)
{
  attroff(A_REVERSE);
  move(2,19);
  switch(usettings_hr.run_level)
  {
    case 0:
      addstr("Verbose");
      break;
    case 1:
      addstr("Normal ");
      break;
    case 2:
      addstr("None   ");
      break;
  }
  move(3,19);
  display_mm_ss(usettings_hr.sleep_time);
  move(4,19);
  display_int(usettings_hr.no_of_warnings);
  move(5,19);
  if(usettings_hr.def_usettings_enab)
    addstr("Yes");
  else
    addstr("No ");
  move(2,48);
  display_int(usettings_hr.clobber_sig);
  move(3,48);
  display_int(usettings_hr.min_free_dialins);

  def_usettings_hl_bar(0);
  clear_area(5,48,5,53);
  tt_field(&(usettings_hr.def_usettings),5);
  def_usettings_hl_bar(1);
  clear_area(5,55,5,59);
  dt_field(&(usettings_hr.def_usettings),5);
  def_usettings_hl_bar(2);
  clear_area(5,61,5,63);
  nice_field(&(usettings_hr.def_usettings),5);
  def_usettings_hl_bar(3);
  clear_area(5,65,5,67);
  cpu_field(&(usettings_hr.def_usettings),5);
  def_usettings_hl_bar(4);
  clear_area(5,69,5,71);
  mem_field(&(usettings_hr.def_usettings),5);
  def_usettings_hl_bar(5);
  coa_field(&(usettings_hr.def_usettings),5);
  def_usettings_hl_bar(6);
  iptd_field(&(usettings_hr.def_usettings),5);
}

/***************************************************************************
*
*  User     Email                                 Total Daily Nic CPU Mem A
*  xxxxxxxx xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx xxx:xx xx:xx xxx xxx xxx x
*
***************************************************************************/
void usettings_hl_bar(unsigned char pl,unsigned char x)
{
  if((usettings_pos_hl == pl) && (usettings_option == x))
    attron(A_REVERSE);
  else
    attroff(A_REVERSE);
}

void d_usettings_entry(int hlpos)
{
  register int entry_no = usettings_pos_real + hlpos;
  register int dis_pos = hlpos + 9;

  usettings_hl_bar(hlpos,0);
  clear_area(dis_pos,2,dis_pos,9);
  usettings_hl_bar(hlpos,1);
  clear_area(dis_pos,11,dis_pos,46);
  usettings_hl_bar(hlpos,2);
  clear_area(dis_pos,48,dis_pos,53);
  usettings_hl_bar(hlpos,3);
  clear_area(dis_pos,55,dis_pos,59);
  usettings_hl_bar(hlpos,4);
  clear_area(dis_pos,61,dis_pos,63);
  usettings_hl_bar(hlpos,5);
  clear_area(dis_pos,65,dis_pos,67);
  usettings_hl_bar(hlpos,6);
  clear_area(dis_pos,69,dis_pos,71);
  usettings_hl_bar(hlpos,7);
  clear_area(dis_pos,73,dis_pos,73);
  usettings_hl_bar(hlpos,8);
  clear_area(dis_pos,75,dis_pos,75);
  if(entry_no < usettings_list_size)
  {
    usettings_hl_bar(hlpos,0);
    move(dis_pos,2);
    if((usettings_list + entry_no)->uid == 65535)
      addnstr("<empty>",8);
    else
    {
      struct passwd *pwuid = getpwuid((usettings_list + entry_no)->uid);
      if(pwuid == NULL)
        addnstr("<defunc>",8);
      else
        addnstr(pwuid->pw_name,8);
    }

    usettings_hl_bar(hlpos,1);
    move(dis_pos,11);
    addnstr((usettings_list + entry_no)->email,STR_EMAIL_LEN);
    usettings_hl_bar(hlpos,2);
    tt_field(usettings_list + entry_no,dis_pos);
    usettings_hl_bar(hlpos,3);
    dt_field(usettings_list + entry_no,dis_pos);
    usettings_hl_bar(hlpos,4);
    nice_field(usettings_list + entry_no,dis_pos);
    usettings_hl_bar(hlpos,5);
    cpu_field(usettings_list + entry_no,dis_pos);
    usettings_hl_bar(hlpos,6);
    mem_field(usettings_list + entry_no,dis_pos);
    usettings_hl_bar(hlpos,7);
    coa_field(usettings_list + entry_no,dis_pos);
    usettings_hl_bar(hlpos,8);
    iptd_field(usettings_list + entry_no,dis_pos);
  }
}

void d_usettings_table(void)
{
  register int hlpos;

  for(hlpos = 0;hlpos < USETTINGS_SCROLL_MAX;hlpos++)
    d_usettings_entry(hlpos);
}

/***************************************************************************
*
* User     A  Total Daily T D C M Real Nice
* xxxxxxxx x xxx:xx xx:xx x x x x xx.xxxxxx
*
***************************************************************************/
void utmp_hl_bar(unsigned char pl,unsigned char x)
{
  if((utmp_pos_hl == pl) && (utmp_option == x))
    attron(A_REVERSE);
  else
    attroff(A_REVERSE);
}

void d_utmp_entry(int hlpos)
{
  register int entry_no = utmp_pos_real + hlpos;
  register int dis_pos = hlpos + 17;
  char str[STR_LEN];

  attroff(A_REVERSE);
  clear_area(dis_pos,2,dis_pos,9);
  move(dis_pos,11);
  addch(' ');
  utmp_hl_bar(hlpos,0);
  clear_area(dis_pos,13,dis_pos,18);
  utmp_hl_bar(hlpos,1);
  clear_area(dis_pos,20,dis_pos,24);
  utmp_hl_bar(hlpos,2);
  move(dis_pos,26);
  addch(' ');
  utmp_hl_bar(hlpos,3);
  move(dis_pos,28);
  addch(' ');
  utmp_hl_bar(hlpos,4);
  move(dis_pos,30);
  addch(' ');
  utmp_hl_bar(hlpos,5);
  move(dis_pos,32);
  addch(' ');
  utmp_hl_bar(hlpos,5);
  clear_area(dis_pos,34,dis_pos,43);

  if(entry_no < utmp_list_size)
  {
    attroff(A_REVERSE);
    move(dis_pos,2);
    addnstr((utmp_list + entry_no)->user,8);

    move(dis_pos,11);
    if((utmp_list + entry_no)->active)
      addch('y');
    else
      addch('n');

    utmp_hl_bar(hlpos,0);
    str_hh_mm((utmp_list + entry_no)->total_time_used,str);
    move(dis_pos,13 + (6 - strlen(str)));
    addstr(str);

    utmp_hl_bar(hlpos,1);
    str_hh_mm((utmp_list + entry_no)->daily_time_used,str);
    move(dis_pos,20 + (5 - strlen(str)));
    addstr(str);

    utmp_hl_bar(hlpos,2);
    move(dis_pos,26);
    display_sint((utmp_list + entry_no)->no_ttime_warnings);

    utmp_hl_bar(hlpos,3);
    move(dis_pos,28);
    display_sint((utmp_list + entry_no)->no_dtime_warnings);

    utmp_hl_bar(hlpos,4);
    move(dis_pos,30);
    display_sint((utmp_list + entry_no)->no_cpu_warnings);

    utmp_hl_bar(hlpos,5);
    move(dis_pos,32);
    display_sint((utmp_list + entry_no)->no_mem_warnings);

    utmp_hl_bar(hlpos,6);
    move(dis_pos,34);
    display_real((utmp_list + entry_no)->nice_factor);
  }
}

void d_utmp_table(void)
{
  register int hlpos;

  for(hlpos = 0;hlpos < USETTINGS_SCROLL_MAX;hlpos++)
    d_utmp_entry(hlpos);
}

/***************************************************************************
*
***************************************************************************/
void display_main_bg(void)
{
  register int x;

  display_topbar();
  move(8,1);
  addstr(" User     Email (*=Flash tty instead)         ");
  addstr(tit_bar1);
  move(16,1);
  addstr(" User     A  Total Daily T D C M Real Nice  ");
  for(x = 9;x < 13;x++)
  {
    move(x,1);
    addch(' ');
    move(x,10);
    addch(' ');
    move(x,47);
    addch(' ');
    move(x,54);
    addch(' ');
    move(x,60);
    addch(' ');
    move(x,64);
    addch(' ');
    move(x,68);
    addch(' ');
    move(x,72);
    addch(' ');
    move(x,74);
    addch(' ');
    move(x,76);
    addch(' ');
  }
  clear_area(13,1,13,76);
  for(x = 17;x < 21;x++)
  {
    move(x,1);
    addch(' ');
    move(x,10);
    addch(' ');
    move(x,12);
    addch(' ');
    move(x,19);
    addch(' ');
    move(x,25);
    addch(' ');
    move(x,27);
    addch(' ');
    move(x,29);
    addch(' ');
    move(x,31);
    addch(' ');
    move(x,33);
    addch(' ');
    move(x,44);
    addch(' ');
  }
  clear_area(21,1,21,44);
  move(4,47);
  addstr(tit_bar1);
  move(5,47);
  addch(' ');
  move(5,54);
  addch(' ');
  move(5,60);
  addch(' ');
  move(5,64);
  addch(' ');
  move(5,68);
  addch(' ');
  move(5,72); 
  addch(' ');
  move(5,74);
  addch(' ');
  move(5,76);
  addch(' ');
  clear_area(6,47,6,76);
  display_hr_titles();
  display_hr_vals();
  d_usettings_table();
  d_utmp_table();
}

/***************************************************************************
*
***************************************************************************/
void clear_usettings_entry(int entry)
{
  (usettings_list + entry)->uid = 65535;
  strcpy((usettings_list + entry)->email,"");
  (usettings_list + entry)->total_time_limit = 0;
  (usettings_list + entry)->daily_time_limit = 0;
  (usettings_list + entry)->nice_factor = 0;
  (usettings_list + entry)->cpu_limit = 0;
  (usettings_list + entry)->mem_limit = 0;
  (usettings_list + entry)->exempt = 0;
}

/***************************************************************************
*
***************************************************************************/
void add_usettings_entry(void)
{
  if(usettings_list_size < pw_table_size)
  {
    clear_usettings_entry(usettings_list_size);
    usettings_list_size++;
  }
  else
    ferr_message("Exceeded entry limit.");
}

/***************************************************************************
*
***************************************************************************/
void edit_usettings(void)
{
  switch(ch)
  {
    case 'A':
      add_usettings_entry();
      d_usettings_table();
      update_screen();
      break;
    case 'D':
      delete_entry(
        usettings_list,
        sizeof(tusettings),
        &usettings_pos_real,&usettings_pos_hl,
        &usettings_list_size,
        &d_usettings_table);
      break;
    case KEY_UP:
      cursor_up(&usettings_pos_real,&usettings_pos_hl,&d_usettings_table);
      break;
    case KEY_DOWN:
      cursor_down(&usettings_pos_real,&usettings_pos_hl,
                  USETTINGS_SCROLL_MAX,usettings_list_size,
                  &d_usettings_table);
      break;
    case KEY_LEFT:
      if(usettings_option > 0) usettings_option--;
      d_usettings_entry(usettings_pos_hl);
      update_screen();
      break;
    case KEY_RIGHT:
      if(usettings_option < 8) usettings_option++;
      d_usettings_entry(usettings_pos_hl);
      update_screen();
      break;
    case 'E':
      if(usettings_list_size)
      {
        register int table_pos = usettings_pos_real + usettings_pos_hl; 
        switch(usettings_option)
        {
          case 2:
            (usettings_list + table_pos)->exempt ^= EXEMPT_TTIME;
            break;
          case 3:
            (usettings_list + table_pos)->exempt ^= EXEMPT_DTIME;
            break;
          case 4:
            (usettings_list + table_pos)->exempt ^= EXEMPT_NICE;
            break;
          case 5:
            (usettings_list + table_pos)->exempt ^= EXEMPT_CPU;
            break;
          case 6:
            (usettings_list + table_pos)->exempt ^= EXEMPT_MEM;
            break;
        }
        d_usettings_table();
        d_utmp_table();
        update_screen();
      }
      break;
    case 10:
      if(usettings_list_size)
      {
        register int dis_pos,table_pos;
        char str[STR_LEN];
        struct passwd *pw_entry;

        dis_pos = usettings_pos_hl + 9;
        table_pos = usettings_pos_real + usettings_pos_hl; 
        switch(usettings_option)
        {
          case 0:
            get_string(dis_pos,2,str,8);
            update_screen();
            if((pw_entry = getpwnam(str)) != NULL)
              (usettings_list + table_pos)->uid = pw_entry->pw_uid;
            d_usettings_table();
            update_screen();
            if(pw_entry == NULL) ferr_message("Not a User. Try again.");
            break;
          case 1:
            get_string(dis_pos,11,(usettings_list + table_pos)->email,
              STR_EMAIL_LEN);
            d_usettings_table();
            update_screen();
            break;
          case 2:
            if(!((usettings_list + table_pos)->exempt & EXEMPT_TTIME))
            {
              (usettings_list + table_pos)->total_time_limit =
                get_hh_mm(dis_pos,48,6);
              d_usettings_table();
              update_screen();
            }
            break;
          case 3:
            if(!((usettings_list + table_pos)->exempt & EXEMPT_DTIME))
            {
              (usettings_list + table_pos)->daily_time_limit =
                get_hh_mm(dis_pos,55,5);
              d_usettings_table();
              update_screen();
            }
            break;
          case 4:
            if(!((usettings_list + table_pos)->exempt & EXEMPT_NICE))
            {
              (usettings_list + table_pos)->nice_factor =
                get_int(dis_pos,61,3);
              range_check(&((usettings_list + table_pos)->nice_factor),-20,19);
              d_usettings_table();
              update_screen();
            }
            break;
          case 5:
            if(!((usettings_list + table_pos)->exempt & EXEMPT_CPU))
            {
              (usettings_list + table_pos)->cpu_limit =
                get_int(dis_pos,65,3);
              range_check(&((usettings_list + table_pos)->cpu_limit),0,100);
              d_usettings_table();
              update_screen();
            }
            break;
          case 6:
            if(!((usettings_list + table_pos)->exempt & EXEMPT_MEM))
            {
              (usettings_list + table_pos)->mem_limit =
                get_int(dis_pos,69,3);
              range_check(&((usettings_list + table_pos)->mem_limit),0,100);
              d_usettings_table();
              update_screen();
            }
            break;
          case 7:
            (usettings_list + table_pos)->exempt ^= EXEMPT_CLOBBERING;
            d_usettings_table();
            update_screen();
            break;
          case 8:
            (usettings_list + table_pos)->exempt ^= INSTANT_PURGE_DT;
            d_usettings_table();
            update_screen();
            break;
        }
      }
      break;
  }
}

/***************************************************************************
*
***************************************************************************/
void edit_utmp(void)
{
  switch(ch)
  {
    case 'D':
      delete_entry(
        utmp_list,
        sizeof(tutmp),
        &utmp_pos_real,&utmp_pos_hl,
        &utmp_list_size,
        &d_utmp_table);
      break;
    case KEY_UP:
      cursor_up(&utmp_pos_real,&utmp_pos_hl,&d_utmp_table);
      break;
    case KEY_DOWN:
      cursor_down(&utmp_pos_real,&utmp_pos_hl,
                  USETTINGS_SCROLL_MAX,utmp_list_size,
                  &d_utmp_table);
      break;
    case KEY_LEFT:
      if(utmp_option > 0) utmp_option--;
      d_utmp_entry(utmp_pos_hl);
      update_screen();
      break;
    case KEY_RIGHT:
      if(utmp_option < 6) utmp_option++;
      d_utmp_entry(utmp_pos_hl);
      update_screen();
      break;
    case 10:
      if(utmp_list_size)
      {
        register int dis_pos = utmp_pos_hl + 17;
        register table_pos = utmp_pos_real + utmp_pos_hl;
        switch(utmp_option)
        {
          case 0:
            (utmp_list + table_pos)->total_time_used =
              get_hh_mm(dis_pos,13,6);
            d_utmp_table();
            update_screen();
            break;
          case 1:
            (utmp_list + table_pos)->daily_time_used =
              get_hh_mm(dis_pos,20,5);
            d_utmp_table();
            update_screen();
            break;
          case 2:
            (utmp_list + table_pos)->no_ttime_warnings =
              get_int(dis_pos,26,1);
            d_utmp_table();
            update_screen();
            break;
          case 3:
            (utmp_list + table_pos)->no_dtime_warnings =
              get_int(dis_pos,28,1);
            d_utmp_table();
            update_screen();
            break;
          case 4:
            (utmp_list + table_pos)->no_cpu_warnings =
              get_int(dis_pos,30,1);
            d_utmp_table();
            update_screen();
            break;
          case 5:
            (utmp_list + table_pos)->no_mem_warnings =
              get_int(dis_pos,32,1);
            d_utmp_table();
            update_screen();
            break;
          case 6:
            (utmp_list + table_pos)->nice_factor =
              get_real(dis_pos,34,10);
            d_utmp_table(); 
            update_screen();
            break;
        }
      }
      break;
  }
}

/***************************************************************************
*
***************************************************************************/
void mainloop(void)
{
  display_main_bg();
  update_screen();
  while(!endedit)
  {
    ch = getch();
    switch(ch)
    {
      case 'Q':
        if(quit_confirm()) endedit = 1;
        break;
      case 'S':
        save_files();
        break;
      case 9:
        if(editmode++ == 8) editmode = 0;
        display_hr_titles();
        update_screen();
        break;
      case 10:
        switch(editmode)
        {
          case 0:
            if(usettings_hr.run_level++ >= 2)
              usettings_hr.run_level = 0;
            display_hr_vals();
            update_screen();
            break;
          case 1:
            usettings_hr.sleep_time = get_mm_ss(3,19,5);
            display_hr_vals();
            update_screen();
            break;
          case 2:
            if(usettings_hr.no_of_warnings++ >= 3)
              usettings_hr.no_of_warnings = 0;
            display_hr_vals();
            update_screen();
            break;
          case 3:
            usettings_hr.def_usettings_enab ^= 1;
            display_hr_vals();
            update_screen();
            break;
          case 4:
            usettings_hr.clobber_sig = get_int(2,48,3);
            display_hr_vals();
            update_screen();
            break;
          case 5:
            usettings_hr.min_free_dialins = get_int(3,48,3);
            display_hr_vals();
            update_screen();
            break;
        }
        break;
    }
    switch(editmode)
    {
      case 6:
        switch(ch)
        {
          case KEY_LEFT:
            if(def_hlpos > 0) def_hlpos--;
            display_hr_vals();
            update_screen();
            break;
          case KEY_RIGHT:
            if(def_hlpos < 6) def_hlpos++;
            display_hr_vals();
            update_screen();
            break;
          case 'E':
            switch(def_hlpos)
            {
              case 0:
                usettings_hr.def_usettings.exempt ^= EXEMPT_TTIME;
                break;
              case 1:
                usettings_hr.def_usettings.exempt ^= EXEMPT_DTIME;
                break;
              case 2:
                usettings_hr.def_usettings.exempt ^= EXEMPT_NICE;
                break;
              case 3:
                usettings_hr.def_usettings.exempt ^= EXEMPT_CPU;
                break;
              case 4:
                usettings_hr.def_usettings.exempt ^= EXEMPT_MEM;
                break;
            }
            display_hr_vals();
            update_screen();
            break;
          case 10:
            switch(def_hlpos)
            {
              case 0:
                if(!(usettings_hr.def_usettings.exempt & EXEMPT_TTIME))
                {
                  usettings_hr.def_usettings.total_time_limit = get_hh_mm(5,48,6);
                  display_hr_vals();
                  update_screen();
                }
                break;
              case 1:
                if(!(usettings_hr.def_usettings.exempt & EXEMPT_DTIME))
                {
                  usettings_hr.def_usettings.daily_time_limit = get_hh_mm(5,55,5);
                  display_hr_vals();
                  update_screen();
                }
                break;
              case 2:
                if(!(usettings_hr.def_usettings.exempt & EXEMPT_NICE))
                {
                  usettings_hr.def_usettings.nice_factor = get_int(5,61,3);
                  range_check(&(usettings_hr.def_usettings.nice_factor),-20,19);
                  display_hr_vals();
                  update_screen();
                }
                break;
              case 3:
                if(!(usettings_hr.def_usettings.exempt & EXEMPT_CPU))
                {
                  usettings_hr.def_usettings.cpu_limit = get_int(5,65,3);
                  range_check(&(usettings_hr.def_usettings.cpu_limit),0,100);
                  display_hr_vals();
                  update_screen();
                }
                break;
              case 4:
                if(!(usettings_hr.def_usettings.exempt & EXEMPT_MEM))
                {
                  usettings_hr.def_usettings.mem_limit = get_int(5,69,3);
                  range_check(&(usettings_hr.def_usettings.mem_limit),0,100);
                  display_hr_vals();
                  update_screen();
                }
                break;
              case 5:
                usettings_hr.def_usettings.exempt ^= EXEMPT_CLOBBERING;
                display_hr_vals();
                update_screen();
                break;
              case 6:
                usettings_hr.def_usettings.exempt ^= INSTANT_PURGE_DT;
                display_hr_vals();
                update_screen();
                break;
            }
            break;
        }
        break;
      case 7:
        edit_usettings();
        break;
      case 8:
        edit_utmp();
        break;
    }
  }
}

/***************************************************************************
*
***************************************************************************/
int main(int argc,char *argv[])
{
  char str[STR_LEN];
  pid_t clobberd_pid;

  printf(WELCOME);
  sprintf(str,"(%s)",argv[0]);
  if(find_process(str,0) > 0)
  {
    printf("Turkey, You are already running clobberedit. Aborting this process..\a\n");
    return -1;
  }
  if((clobberd_pid = find_process("(clobberd)",1)) != -1)
  {
    printf("Sending EDIT mode signal to Clobberd. Please wait 2 seconds..\a\n");
    kill(clobberd_pid,SIGUSR1);
  }
  else
    printf("NB/ 'Clobberd' isn't currently running..\n");
  sleep(2);
  if(load_pw_table())
  {
    if(load_usettings())
    {
      if(load_utmp())
      {
        initscr();
        cbreak();
        noecho();
        keypad(stdscr,TRUE);
        mainloop();
        clear();
        refresh();
        resetterm();
        echo();
        endwin();
        free(utmp_list);
      }
      free(usettings_list);
    }
  }
  if((clobberd_pid = find_process("(clobberd)",1)) != -1)
  {
    printf("Sending MONITOR mode signal to Clobberd. Details will reload next wakeup..\n");
    kill(clobberd_pid,SIGUSR2);
  }
  return 1;
}
